import re

from django.contrib.sites.models import Site
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from redirects.http import (
    HttpResponseGone,
    HttpResponsePermanentRedirect,
    HttpResponseSeeOtherRedirect,
    HttpResponseStrictPermanentRedirect,
    HttpResponseStrictTemporaryRedirect,
    HttpResponseTemporaryRedirect,
)


class Redirect(models.Model):
    TYPE_301 = 301
    TYPE_302 = 302
    TYPE_303 = 303
    TYPE_307 = 307
    TYPE_308 = 308
    TYPE_CHOICES = (
        (TYPE_301, _("301 - Permanent")),
        (TYPE_302, _("302 - Found")),
        (TYPE_303, _("303 - See Other")),
        (TYPE_307, _("307 - Temporary")),
        (TYPE_308, _("308 - Permanent")),
    )

    TYPE_RESPONSE_CLASS = {
        TYPE_301: HttpResponsePermanentRedirect,
        TYPE_302: HttpResponseTemporaryRedirect,
        TYPE_303: HttpResponseSeeOtherRedirect,
        TYPE_307: HttpResponseStrictTemporaryRedirect,
        TYPE_308: HttpResponseStrictPermanentRedirect,
    }

    MATCH_EXACT = "exact"
    MATCH_PREFIX = "prefix"
    MATCH_REGEX = "regex"
    MATCH_CHOICES = (
        (MATCH_EXACT, _("Exact")),
        (MATCH_PREFIX, _("Prefix")),
        (MATCH_REGEX, _("Regex")),
    )

    site = models.ForeignKey(
        Site, blank=True, null=True, on_delete=models.SET_NULL, verbose_name=_("Site")
    )

    old_path = models.CharField(
        db_index=True,
        max_length=255,
        verbose_name=_("Old path"),
        help_text=_(
            "This can be either an absolute path or a regex "
            "(excluding the domain name). "
        ),
    )

    new_path = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("New path"),
        help_text=_(
            "This can be either an absolute path, an absolute URL, or a regex. "
            'If empty a "410 Gone" response will be returned.'
        ),
    )

    match = models.CharField(
        db_index=True,
        max_length=20,
        choices=MATCH_CHOICES,
        default=MATCH_EXACT,
        verbose_name=_("Match"),
        help_text=_("The redirect match condition."),
    )

    type_status_code = models.PositiveSmallIntegerField(
        db_index=True,
        choices=TYPE_CHOICES,
        default=TYPE_301,
        verbose_name=_("Type"),
        help_text=_("The redirect http status code."),
    )

    priority = models.PositiveSmallIntegerField(
        blank=True,
        default=0,
        verbose_name=_("Priority"),
        help_text=_("Higher priority redirects are evaluated first."),
    )

    counter = models.PositiveIntegerField(
        blank=True,
        default=0,
        verbose_name=_("Counter"),
        help_text=_("The redirect requests/responses count."),
    )

    status_code = models.PositiveSmallIntegerField(
        db_index=True,
        blank=True,
        null=True,
        default=None,
        verbose_name=_("Status code"),
    )

    created_at = models.DateTimeField(
        auto_now_add=True,
        editable=False,
        null=True,
        verbose_name=_("Created at"),
    )

    updated_at = models.DateTimeField(
        auto_now=True,
        editable=False,
        null=True,
        verbose_name=_("Updated at"),
    )

    def _get_response_path_with_match_exact(self, path):
        if self.old_path.lower() == path.lower():
            return self.new_path
        return None

    def _get_response_path_with_match_prefix(self, path):
        if path.lower().startswith(self.old_path.lower()):
            return self.new_path
        return None

    def _get_response_path_with_match_regex(self, path):
        try:
            old_path_re = re.compile(self.old_path, re.IGNORECASE)
        except re.error:
            # invalid regex
            return None
        old_path_match = re.match(old_path_re, path)
        if old_path_match:
            new_path_repl = self.new_path.replace("$", "\\")
            return re.sub(old_path_re, new_path_repl, path)
        return None

    def get_response_path(self, path):
        if self.match == Redirect.MATCH_EXACT:
            return self._get_response_path_with_match_exact(path)
        elif self.match == Redirect.MATCH_PREFIX:
            return self._get_response_path_with_match_prefix(path)
        elif self.match == Redirect.MATCH_REGEX:
            return self._get_response_path_with_match_regex(path)
        return None

    def get_response(self, path):
        response_path = self.get_response_path(path)
        if response_path is None:
            return None
        self.counter += 1
        self.save()
        if response_path == "":
            return HttpResponseGone()
        response_class = Redirect.TYPE_RESPONSE_CLASS.get(self.type_status_code)
        return response_class(response_path)

    class Meta:
        app_label = "redirects"
        ordering = ["-priority", "old_path"]
        unique_together = [["site", "old_path"]]
        verbose_name = _("Redirect")
        verbose_name_plural = _("Redirects")

    def __str__(self):
        verbose_name = _("Redirect")
        return force_str(
            f"{verbose_name} {self.type_status_code}: "
            f"{self.old_path} \n---> {self.new_path}"
        )



from django.contrib.sites.shortcuts import get_current_site
from django.db.models import Q

# https://docs.djangoproject.com/en/1.10/topics/http/middleware/#upgrading-pre-django-1-10-style-middleware
from django.utils.deprecation import MiddlewareMixin

from redirects.models import Redirect


class RedirectMiddleware(MiddlewareMixin):
    @staticmethod
    def _find_response(redirects, match, path):
        redirects_list = filter(lambda obj: obj.match == match, redirects)
        for redirect_obj in redirects_list:
            redirect_response = redirect_obj.get_response(path)
            if redirect_response:
                return redirect_response
        return None

    def process_request(self, request):
        current_site = get_current_site(request)
        path = request.get_full_path()

        redirects_qs = Redirect.objects.filter(
            (Q(site=current_site) | Q(site__isnull=True)),
            (Q(match=Redirect.MATCH_EXACT) & Q(old_path__iexact=path))
            | Q(match=Redirect.MATCH_PREFIX)
            | Q(match=Redirect.MATCH_REGEX),
        )
        redirects_list = list(redirects_qs)

        return (
            self._find_response(redirects_list, Redirect.MATCH_EXACT, path)
            or self._find_response(redirects_list, Redirect.MATCH_PREFIX, path)
            or self._find_response(redirects_list, Redirect.MATCH_REGEX, path)
            or None
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("redirects", "0002_add_priority"),
    ]

    operations = [
        migrations.AddField(
            model_name="redirect",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True,
                null=True,
                verbose_name="Created at",
            ),
        ),
        migrations.AddField(
            model_name="redirect",
            name="updated_at",
            field=models.DateTimeField(
                auto_now=True,
                null=True,
                verbose_name="Updated at",
            ),
        ),
    ]



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.CreateModel(
            name="Redirect",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "old_path",
                    models.CharField(
                        db_index=True,
                        help_text=(
                            "This can be either an absolute path or a regex "
                            "(excluding the domain name). "
                        ),
                        max_length=255,
                        verbose_name="Old path",
                    ),
                ),
                (
                    "new_path",
                    models.CharField(
                        blank=True,
                        help_text=(
                            "This can be either an absolute path, "
                            "an absolute URL, or a regex. "
                            'If empty a "410 Gone" response will be returned.'
                        ),
                        max_length=255,
                        verbose_name="New path",
                    ),
                ),
                (
                    "match",
                    models.CharField(
                        choices=[
                            ("exact", "Exact"),
                            ("prefix", "Prefix"),
                            ("regex", "Regex"),
                        ],
                        db_index=True,
                        default="exact",
                        help_text="The redirect match condition.",
                        max_length=20,
                        verbose_name="Match",
                    ),
                ),
                (
                    "type_status_code",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (301, "301 - Permanent"),
                            (302, "302 - Found"),
                            (303, "303 - See Other"),
                            (307, "307 - Temporary"),
                            (308, "308 - Permanent"),
                        ],
                        db_index=True,
                        default=301,
                        help_text="The redirect http status code.",
                        verbose_name="Type",
                    ),
                ),
                (
                    "counter",
                    models.PositiveIntegerField(
                        blank=True,
                        default=0,
                        help_text="The redirect requests/responses count.",
                        verbose_name="Counter",
                    ),
                ),
                (
                    "status_code",
                    models.PositiveSmallIntegerField(
                        blank=True,
                        db_index=True,
                        default=None,
                        null=True,
                        verbose_name="Status code",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="sites.site",
                        verbose_name="Site",
                    ),
                ),
            ],
            options={
                "verbose_name": "Redirect",
                "verbose_name_plural": "Redirects",
                "ordering": ["old_path"],
                "unique_together": {("site", "old_path")},
            },
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("redirects", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="redirect",
            options={
                "ordering": ["-priority", "old_path"],
                "verbose_name": "Redirect",
                "verbose_name_plural": "Redirects",
            },
        ),
        migrations.AddField(
            model_name="redirect",
            name="priority",
            field=models.PositiveSmallIntegerField(
                blank=True,
                default=0,
                help_text="Higher priority redirects are evaluated first.",
                verbose_name="Priority",
            ),
        ),
    ]



__author__ = "Fabio Caccamo"
__copyright__ = "Copyright (c) 2020-present Fabio Caccamo"
__description__ = "django redirects as they should be, with full control."
__email__ = "fabio.caccamo@gmail.com"
__license__ = "MIT"
__title__ = "django-redirects"
__version__ = "0.6.0"



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class RedirectsConfig(AppConfig):
    name = "redirects"
    verbose_name = _("Redirects")
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        pass



from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse

from redirects.models import Redirect


class AdminTestCase(TestCase):
    """
    This class describes an admin test case.
    """

    def setUp(self):
        self._superuser = User.objects.create_superuser(
            "superuser", "superuser@django-redirects.test", "test"
        )
        self._superuser.save()
        self._client = Client()
        login = self._client.login(username="superuser", password="test")
        assert login
        self._redirect = Redirect.objects.create(
            old_path="/obsolete/page-301/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
        )

    def test_admin_urls(self):
        url = reverse("admin:redirects_redirect_add")
        response = self._client.get(url)
        self.assertEqual(response.status_code, 200)

        url = reverse("admin:redirects_redirect_changelist")
        response = self._client.get(url)
        self.assertEqual(response.status_code, 200)

        url = reverse("admin:redirects_redirect_change", args=(self._redirect.id,))
        response = self._client.get(url)
        self.assertEqual(response.status_code, 200)

        url = reverse("admin:redirects_redirect_delete", args=(self._redirect.id,))
        response = self._client.get(url)
        self.assertEqual(response.status_code, 200)



from django.contrib import admin
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from redirects.models import Redirect


@admin.register(Redirect)
class RedirectAdmin(admin.ModelAdmin):
    def redirect_display(self, obj):
        gone = _("(410 Gone)") if obj.new_path == "" else ""
        html = f"""
            <span style="line-height: 16px;">
                <span style="display: block; white-space: nowrap; font-weight: normal;">
                    <small>{obj.old_path}</small>
                </span>
                <span style="display: block; white-space: nowrap;">
                    <span style="color: rgba(0, 0, 0, 0.4);">&searr; {gone}</span> {obj.new_path}
                </span>
            </span>
            """.strip()  # noqa: E501
        html = mark_safe(html)
        return html

    redirect_display.short_description = _("Redirect")

    def test_display(self, obj):
        css = """
            font-weight: normal;
            font-size: 20px;
            line-height: 1em;
            display: inline-block;
            padding: 0px 7px 7px 7px;
            """.strip()
        if obj.new_path == "" or obj.match == Redirect.MATCH_REGEX:
            css += """
                opacity: 0.2;
                pointer-events: none;
                """.strip()
        html = f"""
            <a href="{obj.old_path}" target="_blank" style="{css}">&nearr;</a>
            """.strip()
        html = mark_safe(html)
        return html

    test_display.short_description = _("Test")

    list_display = (
        "redirect_display",
        "match",
        "type_status_code",
        "priority",
        "counter",
        "status_code",
        "test_display",
        "created_at",
        "updated_at",
    )
    list_display_links = ("redirect_display",)
    list_editable = (
        "match",
        "type_status_code",
        "priority",
    )
    list_filter = (
        "site",
        "match",
        "type_status_code",
        "status_code",
    )
    search_fields = (
        "old_path",
        "new_path",
    )
    save_on_top = True



from django.http import HttpResponseGone, HttpResponsePermanentRedirect  # noqa: F401
from django.http import (  # noqa: F401
    HttpResponseRedirect as HttpResponseTemporaryRedirect,
)
from django.http.response import HttpResponseRedirectBase


class HttpResponseSeeOtherRedirect(HttpResponseRedirectBase):
    status_code = 303


class HttpResponseStrictTemporaryRedirect(HttpResponseRedirectBase):
    status_code = 307


class HttpResponseStrictPermanentRedirect(HttpResponseRedirectBase):
    status_code = 308



from django.http import HttpResponseNotFound
from django.test import Client, RequestFactory, TestCase
from django.urls import reverse

from redirects.models import Redirect


class ModelsTestCase(TestCase):
    """
    This class describes a models test case.
    """

    def setUp(self):
        self._client = Client()
        self._request_factory = RequestFactory()

    def test_redirect_type_choices(self):
        self.assertEqual(Redirect.TYPE_301, 301)
        self.assertEqual(Redirect.TYPE_302, 302)
        self.assertEqual(Redirect.TYPE_303, 303)
        self.assertEqual(Redirect.TYPE_307, 307)
        self.assertEqual(Redirect.TYPE_308, 308)

    def test_redirect_match_choices(self):
        self.assertEqual(Redirect.MATCH_EXACT, "exact")
        self.assertEqual(Redirect.MATCH_PREFIX, "prefix")
        self.assertEqual(Redirect.MATCH_REGEX, "regex")

    def test_redirect_301_url_and_status_code(self):
        Redirect.objects.create(
            old_path="/obsolete/page-301/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
        )
        response = self._client.get("/obsolete/page-301/")
        self.assertEqual(response.url, reverse("redirected-301"))
        self.assertEqual(response.status_code, 301)

    def test_redirect_302_url_and_status_code(self):
        Redirect.objects.create(
            old_path="/obsolete/page-302/",
            new_path=reverse("redirected-302"),
            type_status_code=Redirect.TYPE_302,
        )
        response = self._client.get("/obsolete/page-302/")
        self.assertEqual(response.url, reverse("redirected-302"))
        self.assertEqual(response.status_code, 302)

    def test_redirect_303_url_and_status_code(self):
        Redirect.objects.create(
            old_path="/obsolete/page-303/",
            new_path=reverse("redirected-303"),
            type_status_code=Redirect.TYPE_303,
        )
        response = self._client.get("/obsolete/page-303/")
        self.assertEqual(response.url, reverse("redirected-303"))
        self.assertEqual(response.status_code, 303)

    def test_redirect_307_url_and_status_code(self):
        Redirect.objects.create(
            old_path="/obsolete/page-307/",
            new_path=reverse("redirected-307"),
            type_status_code=Redirect.TYPE_307,
        )
        response = self._client.get("/obsolete/page-307/")
        self.assertEqual(response.url, reverse("redirected-307"))
        self.assertEqual(response.status_code, 307)

    def test_redirect_308_url_and_status_code(self):
        Redirect.objects.create(
            old_path="/obsolete/page-308/",
            new_path=reverse("redirected-308"),
            type_status_code=Redirect.TYPE_308,
        )
        response = self._client.get("/obsolete/page-308/")
        self.assertEqual(response.url, reverse("redirected-308"))
        self.assertEqual(response.status_code, 308)

    def test_redirect_match_exact(self):
        Redirect.objects.create(
            old_path="/obsolete-match-exact/page-301/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
            match=Redirect.MATCH_EXACT,
        )
        response = self._client.get("/obsolete-match-exact/page-301/")
        self.assertEqual(response.url, reverse("redirected-301"))

    def test_redirect_match_exact_with_wrong_url(self):
        Redirect.objects.create(
            old_path="/obsolete-match-exact/page-301/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
            match=Redirect.MATCH_EXACT,
        )
        response = self._client.get("/obsolete-match-exact/page-30X/")
        self.assertTrue(isinstance(response, HttpResponseNotFound))

    def test_redirect_match_prefix(self):
        Redirect.objects.create(
            old_path="/obsolete-match-prefix/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
            match=Redirect.MATCH_PREFIX,
        )
        response = self._client.get("/obsolete-match-prefix/page-301/")
        self.assertEqual(response.url, reverse("redirected-301"))

    def test_redirect_match_prefix_with_wrong_url(self):
        Redirect.objects.create(
            old_path="/ooobsolete-match-prefix/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
            match=Redirect.MATCH_PREFIX,
        )
        response = self._client.get("/obsolete-match-prefix/page-301/")
        self.assertTrue(isinstance(response, HttpResponseNotFound))

    def test_redirect_hits_counter(self):
        redirect_obj = Redirect.objects.create(
            old_path="/obsolete/page-counter/",
            new_path=reverse("redirected-301"),
            type_status_code=Redirect.TYPE_301,
        )
        for _ in range(0, 5):
            self._client.get("/obsolete/page-counter/")
        self.assertEqual(Redirect.objects.get(pk=redirect_obj.pk).counter, 5)
